/*
 * Copyright (C) 2002-2007 The Warp Rogue Team
 * Part of the Warp Rogue Project
 *
 * This software is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License.
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY.
 *
 * See the license.txt file for more details.
 */

/*
 * Module Name: DataFile
 * Description: -
 */

#define Uses_Ui
#define Uses_ProgramManager
#define Uses_Util
#define Uses_Equipment
#define Uses_Career
#define Uses_Terrain
#define Uses_Object
#define Uses_Stats
#define Uses_Faction
#define Uses_Perks
#define Uses_Inventory
#define Uses_Npc
#define Uses_Character
#define Uses_Stats
#define Uses_Psychic
#define Uses_Options
#define Uses_WorldGenerator
/* hack - fix later */
#define Uses_Perception
#include "mheader.h"

#include "rdb.h"

#include "datafile.h"

#include "files.h"


static char *		data_file_entry_description(char *, const char *, const char *);

static void		load_keybindings(void);
static void		load_options(void);

static void		load_terrain_files(void);
static void		load_terrain_file(const char *);

static void		load_object_files(void);
static void		load_object_file(const char *);

static void		load_career_files(void);
static void		load_career_file(const char *);

static void		load_npc_files(void);
static void		load_npc_file(const char *);
static void		load_npc_pp(CHARACTER *);

static void		load_spawn_file(void);

static void		read_symbol_field(SYMBOL *);
static void		read_dynamic_color_field(DYNAMIC_COLOR *);
static void		read_stat_field(CHARACTER *, const char *);
static void		read_firing_mode_field(OBJECT_DATA *);
static void		read_dice_roll_field(DICE_ROLL *);
static OBJECT *		read_object_field(void);
static void		read_career_field(CHARACTER *);
static void		read_description_field(char *, const char *, const char *);
static void		read_key_field(KEY_CODE *);
static void		read_option_value_field(OPTION_VALUE *);
static void		read_boolean_field(bool *);
static void		read_ai_state_field(CHARACTER *);
static void		read_gender_field(GENDER *);
static void		read_name_field(char *, unsigned int);



/*
 * object files
 */
static const char *	ObjectFile[N_OBJECT_FILES] = {

	FILE_CCW_OBJECTS,
	FILE_RCW_OBJECTS,
	FILE_ARMOUR_OBJECTS,
	FILE_DRUG_OBJECTS,
	FILE_ENVIRONMENT_OBJECTS,
	FILE_PSYCHIC_BOLTS,
	FILE_MISC_OBJECTS

};





/*
 * loads the data files
 */
void load_data_files(void)
{

	load_keybindings();

	load_options();

	load_terrain_files();

	load_object_files();

	load_career_files();

	load_npc_files();

	load_spawn_file();

}



/*
 * returns the description of a terrain
 */
char * data_file_terrain_description(char *description, const TERRAIN *terrain)
{
	char file_path[FILE_PATH_SIZE];

	data_path(file_path, DIR_TERRAIN, FILE_TERRAIN);

	data_file_entry_description(description, file_path,
		terrain_static_data(terrain)->name
	);

	return description;
}



/*
 * returns the description of an object
 */
char * data_file_object_description(char *description, const OBJECT *object)
{
	char file_path[FILE_PATH_SIZE];
	int file_index;
	const OBJECT_DATA *object_data;

	object_data = object_static_data(object);

	if (object_data->type == OTYPE_NIL ||
		object_data->type > OTYPE_ENVIRONMENT) {

		file_index = N_OBJECT_FILES - 1;
	} else {
		file_index = object_data->type - 1;
	}

	data_path(file_path, DIR_OBJECTS, ObjectFile[file_index]);

	data_file_entry_description(description, file_path,
		object_data->name
	);

	return description;
}



/*
 * returns the description of a career
 */
char * data_file_career_description(char *description, const CAREER *career)
{
	char file_path[FILE_PATH_SIZE];

	if (career->type == CAREER_TYPE_BASIC) {
		data_path(file_path, DIR_CHARACTERS, FILE_BASIC_CAREERS);
	} else if (career->type == CAREER_TYPE_ADVANCED) {
		data_path(file_path, DIR_CHARACTERS, FILE_ADVANCED_CAREERS);
	} else {
		die("*** CORE ERROR *** invalid career description call");
	}

	data_file_entry_description(description, file_path, career->name);

	return description;
}



/*
 * returns the description of a character
 */
char * data_file_character_description(char *description, const CHARACTER *character)
{
	char file_path[FILE_PATH_SIZE];

	data_path(file_path, DIR_CHARACTERS, FILE_NPCS);

	data_file_entry_description(description, file_path, character->name);

	if (ZERO_LENGTH_STRING(description) && character->career != CAREER_NIL) {
		CAREER *career;

		career = get_career_pointer(character->career);
		career_description(description, career);
	}

	return description;
}



/*
 * returns the description of a data file entry
 */
static char * data_file_entry_description(char *description, const char *file_path,
	const char *entry_name
)
{
	bool entry_reached = false;

	rdb_open(file_path, RDB_READ);

	CLEAR_STRING(description);

	while (rdb_next_field()) {
		const char *field_name;

                field_name = rdb_field_name();

		if (STR_EQ(field_name, "NAME")) {

			if (entry_reached) break;

			if (STR_EQ(rdb_data_token(RDB_NEXT_TOKEN),
				entry_name)) {

				entry_reached = true;
			}
			continue;
		}

		if (entry_reached && STR_EQ(field_name, "DESCRIPTION")) {
			read_description_field(description, file_path,
				entry_name
			);
		}

	}

	rdb_close();

	return description;
}



/*
 * loads the keybindings
 */
static void load_keybindings(void)
{
	char file_path[FILE_PATH_SIZE];
	COMMAND command = CM_NIL;

	data_path(file_path, DIR_CONFIG, KEYBINDINGS_FILE);

	rdb_open(file_path, RDB_READ);

        while (rdb_next_field()) {
		const char *field_name;

                field_name = rdb_field_name();

		if (STR_EQ(field_name, "COMMAND")) {

			command = name_to_command(
				rdb_data_token(RDB_NEXT_TOKEN)
			);

		} else if (STR_EQ(field_name, "KEY")) {
			KEY_CODE key;

			read_key_field(&key);

			set_command_key(command, key);

		} else {

			die("*** CORE ERROR *** invalid field " \
				"name: %s (%s)",
				field_name,
				file_path
			);
		}

	}

	rdb_close();

}



/*
 * loads the options file
 */
static void load_options(void)
{
	char file_path[FILE_PATH_SIZE];

	data_path(file_path, DIR_CONFIG, OPTIONS_FILE);

	rdb_open(file_path, RDB_READ);

        while (rdb_next_field()) {
		const char *field_name;
		OPTION_VALUE option_value;

                field_name = rdb_field_name();

		read_option_value_field(&option_value);

		if (STR_EQ(field_name, "FULLSCREEN")) {

			option_set_value(OPT_FULLSCREEN,
				option_value
			);

		} else if (STR_EQ(field_name, "BRIGHTNESS")) {

			option_set_value(OPT_BRIGHTNESS,
				option_value
			);

		} else if (STR_EQ(field_name, "CENTER_ON_PLAYER")) {

			option_set_value(OPT_CENTER_ON_PLAYER,
				option_value
			);

		} else {

			die("*** CORE ERROR *** invalid field name: " \
				"%s (%s)",
				field_name,
				file_path
			);
		}
	}

	rdb_close();
}



/*
 * loads the terrain files
 */
static void load_terrain_files(void)
{
	char file_path[FILE_PATH_SIZE];

	data_path(file_path, DIR_TERRAIN, FILE_TERRAIN);

	load_terrain_file(file_path);

	terrain_box_optimize();
}



/*
 * loads a terrain file
 */
static void load_terrain_file(const char *file_path)
{
	char description[DESCRIPTION_SIZE];
	TERRAIN_DATA *terrain_data = NULL;

	rdb_open(file_path, RDB_READ);

        while (rdb_next_field()) {
		const char *field_name;

                field_name = rdb_field_name();

		if (STR_EQ(field_name, "NAME")) {

			terrain_data = terrain_box_new_terrain();

			CLEAR_STRING(description);

			read_name_field(terrain_data->name,
				TERRAIN_NAME_SIZE
			);

		} else if (STR_EQ(field_name, "SYMBOL")) {

			read_symbol_field(&terrain_data->symbol);

		} else if (STR_EQ(field_name, "COLOR")) {

			read_dynamic_color_field(&terrain_data->color);

		} else if (STR_EQ(field_name, "ATTRIBUTE")) {
			TERRAIN_ATTRIBUTE attribute;

			attribute = name_to_terrain_attribute(
				rdb_data_token(RDB_NEXT_TOKEN)
			);

			terrain_data->attribute[attribute] = true;

		} else if (STR_EQ(field_name, "DESCRIPTION")) {

			read_description_field(description, file_path,
				terrain_data->name
			);

		} else {

			die("*** CORE ERROR *** invalid field name: " \
				"%s (%s)",
				field_name,
				file_path
			);
		}


        }

	rdb_close();
}



/*
 * loads the object files
 */
static void load_object_files(void)
{
	char file_path[FILE_PATH_SIZE];
	int i;

	for (i = 0; i < N_OBJECT_FILES; i++) {

		data_path(file_path, DIR_OBJECTS, ObjectFile[i]);

		load_object_file(file_path);
	}

	object_box_optimize();
}



/*
 * loads an object file
 */
static void load_object_file(const char *file_path)
{
	char description[DESCRIPTION_SIZE];
	OBJECT_DATA *object_data = NULL;

	rdb_open(file_path, RDB_READ);

        while (rdb_next_field()) {
		const char *field_name;

                field_name = rdb_field_name();

		if (STR_EQ(field_name, "NAME")) {

			object_data = object_box_new_object();

			CLEAR_STRING(description);

			read_name_field(object_data->name,
				OBJECT_NAME_SIZE
			);

		} else if (STR_EQ(field_name, "SYMBOL")) {

			read_symbol_field(&object_data->symbol);

		} else if (STR_EQ(field_name, "COLOR")) {

			read_dynamic_color_field(&object_data->color);

		} else if (STR_EQ(field_name, "TYPE")) {

			object_data->type = name_to_object_type(
				rdb_data_token(RDB_NEXT_TOKEN)
			);

		} else if (STR_EQ(field_name, "SUBTYPE")) {

			object_data->subtype = name_to_object_subtype(
				rdb_data_token(RDB_NEXT_TOKEN)
			);

		} else if (STR_EQ(field_name, "CONDITION")) {

			object_data->condition_max = atoi(
				rdb_data_token(RDB_NEXT_TOKEN)
			);

		} else if (STR_EQ(field_name, "VALUE")) {

			object_data->value = atoi(
				rdb_data_token(RDB_NEXT_TOKEN)
			);

		} else if (STR_EQ(field_name, "BUYABLE")) {

			read_boolean_field(&object_data->buyable);

		} else if (STR_EQ(field_name, "WEIGHT")) {

			object_data->weight = atoi(
				rdb_data_token(RDB_NEXT_TOKEN)
			);

		} else if (STR_EQ(field_name, "DAMAGE")) {

			read_dice_roll_field(&object_data->damage);

		} else if (STR_EQ(field_name, "ARMOUR_VALUE")) {

			object_data->armour_value = atoi(
				rdb_data_token(RDB_NEXT_TOKEN)
			);

		} else if (STR_EQ(field_name, "ACCURACY")) {

			object_data->firing_mode.accuracy = atoi(
				rdb_data_token(RDB_NEXT_TOKEN)
			);

		} else if (STR_EQ(field_name, "SHOTS")) {

			object_data->charge_max = atoi(
				rdb_data_token(RDB_NEXT_TOKEN)
			);

		} else if (STR_EQ(field_name, "FIRING_MODE")) {

			read_firing_mode_field(object_data);

		} else if (STR_EQ(field_name, "RELOAD")) {

			object_data->reload = atoi(
				rdb_data_token(RDB_NEXT_TOKEN)
			);

		} else if (STR_EQ(field_name, "ATTRIBUTE")) {
			OBJECT_ATTRIBUTE attribute;

			attribute = name_to_object_attribute(
				rdb_data_token(RDB_NEXT_TOKEN)
			);

			object_data->attribute[attribute] = true;

		} else if (STR_EQ(field_name, "REACH")) {

			object_data->reach = atoi(
				rdb_data_token(RDB_NEXT_TOKEN)
			);

		} else if (STR_EQ(field_name, "PARRY_PENALTY")) {

			object_data->parry_penalty = atoi(
				rdb_data_token(RDB_NEXT_TOKEN)
			);

		} else if (STR_EQ(field_name, "DESCRIPTION")) {

			read_description_field(description, file_path,
				object_data->name
			);

		} else {

			die("*** CORE ERROR *** invalid field name: %s (%s)",
				field_name,
				file_path
			);
		}


        }

	rdb_close();
}



/*
 * loads the career files
 */
static void load_career_files(void)
{
	char file_path[FILE_PATH_SIZE];

	data_path(file_path, DIR_CHARACTERS, FILE_BASIC_CAREERS);
	load_career_file(file_path);

	data_path(file_path, DIR_CHARACTERS, FILE_ADVANCED_CAREERS);
	load_career_file(file_path);

	career_box_optimize();
}



/*
 * loads a career file
 */
static void load_career_file(const char *file_path)
{
	char description[DESCRIPTION_SIZE];
	CAREER *career = NULL;
	int career_exit_index = 0;

	rdb_open(file_path, RDB_READ);

        while (rdb_next_field()) {
		const char *field_name;

                field_name = rdb_field_name();

		if (STR_EQ(field_name, "NAME")) {

			career = career_box_new_career();

			career_exit_index = 0;

			CLEAR_STRING(description);

			read_name_field(career->name,
				CAREER_NAME_SIZE
			);

		} else if (STR_EQ(field_name, "TYPE")) {

			career->type = name_to_career_type(
				rdb_data_token(RDB_NEXT_TOKEN)
			);

		} else if (STR_EQ(field_name, "COLOR")) {

			career->color = name_to_color(
				rdb_data_token(RDB_NEXT_TOKEN)
			);

		} else if (is_short_stat_name(field_name)) {
			STAT i;

			i = name_to_stat(field_name, true);

			career->max_stat_advance[i] = atoi(
				rdb_data_token(RDB_NEXT_TOKEN)
			);

		} else if (STR_EQ(field_name, "PERK")) {
			PERK perk;

			perk = name_to_perk(
				rdb_data_token(RDB_NEXT_TOKEN)
			);

			career->perk[perk] = true;


		} else if (STR_EQ(field_name, "OBJECT")) {
			OBJECT *object;

			object = read_object_field();

			list_add(career->trappings, object);

		} else if (STR_EQ(field_name, "EXIT")) {

			strcpy(career->exit[career_exit_index],
				rdb_data_token(RDB_NEXT_TOKEN)
			);

			++career_exit_index;

		} else if (STR_EQ(field_name, "DESCRIPTION")) {

			read_description_field(description, file_path,
				career->name
			);

		} else {

			die("*** CORE ERROR *** invalid field name: %s (%s)",
				field_name,
				file_path
			);

		}


        }

        rdb_close();
}



/*
 * loads the NPC files
 */
static void load_npc_files(void)
{
	char file_path[FILE_PATH_SIZE];

	data_path(file_path, DIR_CHARACTERS, FILE_NPCS);

	load_npc_file(file_path);

	npc_box_optimize();
}



/*
 * loads a NPC file
 */
static void load_npc_file(const char *file_path)
{
	char description[DESCRIPTION_SIZE];
	CHARACTER *npc = NULL;

        rdb_open(file_path, RDB_READ);

	while (rdb_next_field()) {
		const char *field_name;

                field_name = rdb_field_name();

		if (STR_EQ(field_name, "NAME")) {

			if (npc != NULL) {

				load_npc_pp(npc);
			}

			npc = npc_box_new_npc();

			CLEAR_STRING(description);

			read_name_field(npc->name,
				CHARACTER_NAME_SIZE
			);

		} else if (STR_EQ(field_name, "CAREER")) {

			read_career_field(npc);

		} else if (STR_EQ(field_name, "HOSTILE")) {

			read_boolean_field(&npc->is_hostile);

		} else if (STR_EQ(field_name, "GENDER")) {

			read_gender_field(&npc->gender);

		} else if (STR_EQ(field_name, "SYMBOL")) {

			read_symbol_field(&npc->symbol);

		} else if (STR_EQ(field_name, "COLOR")) {

			npc->color = name_to_color(
				rdb_data_token(RDB_NEXT_TOKEN)
			);

		} else if (is_short_stat_name(field_name)) {

			read_stat_field(npc, field_name);

		} else if (STR_EQ(field_name, "OBJECT")) {
			OBJECT *object;

			object = read_object_field();

			inventory_add(npc, object);

		} else if (STR_EQ(field_name, "PERK")) {
			PERK perk;

			perk = name_to_perk(
				rdb_data_token(RDB_NEXT_TOKEN)
			);

			gain_perk(npc, perk);

		} else if (STR_EQ(field_name, "PSYCHIC_POWER")) {
			PSY_POWER power;

			power = name_to_psy_power(
				rdb_data_token(RDB_NEXT_TOKEN)
			);

			npc->psy_power[power] = true;

		} else if (STR_EQ(field_name, "STATUS")) {
			CHARACTER_FLAG flag;

			flag = name_to_character_flag(
				rdb_data_token(RDB_NEXT_TOKEN)
			);

			character_set_flag(npc, flag);

		} else if (STR_EQ(field_name, "RECRUITABLE")) {

			read_boolean_field(&npc->recruitable);

		} else if (STR_EQ(field_name, "AI_DEFAULT_STATE")) {

			read_ai_state_field(npc);

		} else if (STR_EQ(field_name, "DESCRIPTION")) {

			read_description_field(description, file_path,
				npc->name
			);

		} else {

			die("*** CORE ERROR *** invalid field name: %s (%s)",
				field_name,
				file_path
			);
		}
        }

	if (npc != NULL) {

		load_npc_pp(npc);
	}

	rdb_close();
}



/*
 * NPC loading post-processor
 */
static void load_npc_pp(CHARACTER *npc)
{

	equip_objects(npc);

	if (npc->perk[PK_STEALTH]) {

		activate_stealth(npc);
	}
}



/*
 * loads the spawn file
 */
static void load_spawn_file(void)
{
	char file_path[FILE_PATH_SIZE];
	WORLD_COORD hive_level;

	hive_level = 0;

	data_path(file_path, DIR_CHARACTERS, FILE_SPAWN);

	rdb_open(file_path, RDB_READ);

	while (rdb_next_field()) {
		const char *field_name;

                field_name = rdb_field_name();

		if (STR_EQ(field_name, "LEVEL")) {

			hive_level = atoi(
				rdb_data_token(RDB_NEXT_TOKEN)
			);

		} else if (STR_EQ(field_name, "NPC")) {

			spawn_list_add(hive_level,
				rdb_data_token(RDB_NEXT_TOKEN)
			);

		} else if (STR_EQ(field_name, "PC")) {

			set_pc_spawn_position(hive_level);

		} else {

			die("*** CORE ERROR *** invalid field name: %s (%s)",
				field_name,
				file_path
			);

		}

	}

	rdb_close();
}



/*
 * reads symbol field data
 */
static void read_symbol_field(SYMBOL *symbol)
{

	if (strlen(rdb_data_token(0)) > 1) {
		*symbol = atoi(&(rdb_data_token(0)[1]));
	} else {
		*symbol = rdb_data_token(0)[0];
	}
}



/*
 * reads dynamic color field data
 */
static void read_dynamic_color_field(DYNAMIC_COLOR *color)
{

	color->primary = name_to_color(rdb_data_token(0));

	if (rdb_optional_data_token(1) == NULL) {
		color->secondary = C_NIL;
		color->frequency = 0;
		return;
	}

	color->secondary = name_to_color(rdb_data_token(1));
	color->frequency = atoi(rdb_data_token(2));
}



/*
 * reads stat field data
 */
static void read_stat_field(CHARACTER *character, const char *field_name)
{
	STAT i;

	i = name_to_stat(field_name, true);

	character->stat[i].current =
	character->stat[i].total = atoi(
		rdb_data_token(0)
	);

	if (rdb_optional_data_token(1) != NULL) {

		character->stat[i].advance =
			atoi(rdb_data_token(1));
	}
}



/*
 * reads firing mode field data
 */
static void read_firing_mode_field(OBJECT_DATA *object_data)
{
	FIRING_MODE firing_mode;

	firing_mode = name_to_firing_mode(
		rdb_data_token(RDB_NEXT_TOKEN)
	);

	object_data->firing_mode.has[firing_mode] = true;

	if (firing_mode == FMODE_S) {

		return;

	} else if (firing_mode == FMODE_SA) {

		object_data->firing_mode.min_sa_shots =
			atoi(rdb_data_token(RDB_NEXT_TOKEN));
		object_data->firing_mode.max_sa_shots =
			atoi(rdb_data_token(RDB_NEXT_TOKEN));

		return;

	} else if (firing_mode == FMODE_A) {

		object_data->firing_mode.a_shots =
			atoi(rdb_data_token(RDB_NEXT_TOKEN));
	}

}



/*
 * reads dice roll field data
 */
static void read_dice_roll_field(DICE_ROLL *dice_roll)
{

	dice_roll->n_dice = atoi(rdb_data_token(RDB_NEXT_TOKEN));

	dice_roll->n_sides = atoi(rdb_data_token(RDB_NEXT_TOKEN));

	if (rdb_optional_data_token(2) == NULL) {
		dice_roll->modifier = 0;
		return;
	}

	dice_roll->modifier = atoi(rdb_data_token(2));
}



/*
 * reads object field data
 */
static OBJECT * read_object_field(void)
{
	OBJECT *object;

	object = object_create(rdb_data_token(0));

	if (rdb_optional_data_token(1) == NULL) {
		return object;
	}

	object->charge = atoi(rdb_data_token(1));

	return object;
}



/*
 * reads career field data
 */
static void read_career_field(CHARACTER *character)
{
	CAREER *career;

	character->career = name_to_career_index(
		rdb_data_token(RDB_NEXT_TOKEN)
	);

	career = get_career_pointer(character->career);
	character->color = career->color;

}



/*
 * reads description field data
 */
static void read_description_field(char *description, const char *file_path, const char *entry_name)
{
	char text_buffer[TEXT_BUFFER_SIZE];

	strcpy(text_buffer, description);
	strcat(text_buffer, rdb_data_token(RDB_NEXT_TOKEN));

	if (strlen(text_buffer) >= DESCRIPTION_SIZE) {

		die("*** CORE ERROR *** description of %s in %s is too long " \
			"(limit: %d characters)",
			entry_name, file_path, DESCRIPTION_SIZE - 1
		);
	}

	strcpy(description, text_buffer);
}



/*
 * reads key field data
 */
static void read_key_field(KEY_CODE *key)
{

	*key = name_to_key_code(rdb_data_token(RDB_NEXT_TOKEN));
}



/*
 * reads option value field data
 */
static void read_option_value_field(OPTION_VALUE *option_value)
{

	*option_value = (OPTION_VALUE)(
		atoi(rdb_data_token(RDB_NEXT_TOKEN))
	);
}



/*
 * reads boolean field data
 */
static void read_boolean_field(bool *data)
{

	if (STR_EQ(rdb_data_token(0), "Yes")) {

		*data = true;

	} else if (STR_EQ(rdb_data_token(0), "No")) {

		*data = false;

	} else {

		die("*** CORE ERROR *** invalid data in " \
			"boolean field: %s", rdb_data_token(0)
		);
	}

}



/*
 * reads AI state field data
 */
static void read_ai_state_field(CHARACTER *character)
{

	if (STR_EQ(rdb_data_token(0), "Wait")) {

		character->ai.state =
			character->ai.default_state =
			character->ai.original_state =
			AI_STATE_WAIT;

	} else {

		die("*** CORE ERROR *** invalid data in " \
			"AI state field: %s", rdb_data_token(0)
		);
	}

}



/*
 * reads gender field data
 */
static void read_gender_field(GENDER *gender)
{

	switch (rdb_data_token(0)[0]) {

	case 'M':
		*gender = GENDER_MALE;
		break;
	case 'F':
		*gender = GENDER_FEMALE;
		break;
	case 'N':
		*gender = GENDER_NEUTER;
		break;
	default:
		die("*** CORE ERROR *** invalid data in gender field: %c",
			rdb_data_token(0)[0]
		);
	}

}


/*
 * reads name field data
 */
static void read_name_field(char *name, unsigned int n)
{

	if (strlen(rdb_data_token(0)) < n) {
		strcpy(name, rdb_data_token(0));
		return;
	}

	die("*** CORE ERROR *** name too long: %s " \
		"(limit: %d characters)",
		rdb_data_token(0), n - 1
	);
}



